import math
import random
import collections

maxSimulationTime = 10

class Node:
    def __init__(self, position, A):
        
        #Initialize node variables
        self.position = position
        self.num_collisions = 0
        self.wait_collisions = 0
        self.MAX_COLLISIONS = 10
        
        #Create a dequeue containing the packets
        self.queue = collections.deque(self.make_queue(A))
        
    def make_queue(self, A):
        
        #Generate packets based on uniform exponential distribution (Lab1)
        packets = []
        packet_time_ctr = 0
        
        while packet_time_ctr <= maxSimulationTime:
            packet_time_ctr += get_exp_rv(A)
            packets.append(packet_time_ctr)
        
        #Sort packets based on time
        return sorted(packets)
    
    def drop_packet(self):
        self.queue.popleft()
        self.num_collisions = 0
        self.wait_collisions = 0
        
    def get_exp_backoff(self, R, i):
        num = random.random()*(pow(2,i)-1)
        num_backoff = num*512/float(R)
        return num_backoff
    
    def transmission_failed(self, R):
        self.num_collisions += 1
        if self.num_collisions > self.MAX_COLLISIONS:
            return self.drop_packet()
        
        backoff_time = self.queue[0] + self.get_exp_backoff(R, self.num_collisions)
        
        for i in range(len(self.queue)):
            if backoff_time >= self.queue[i]:
                self.queue[i] = backoff_time
            else:
                break
            
    def transmission_success(self):
        self.num_collisions = 0
        self.wait_collisions = 0
        
    def medium_busy_np(self,R):
        self.wait_collisions += 1
        if self.wait_collisions > self.MAX_COLLISIONS:
            return self.drop_packet()
        
        backoff_time = self.queue[0] + self.get_exp_backoff(R, self.wait_collisions)
        
        for i in range(len(self.queue)):
            if backoff_time >= self.queue[i]:
                self.queue[i] = backoff_time
            else:
                break
       
def get_exp_rv(value):
    
    #Exponential Random Variable generator 
    uniform_rv = 1-random.uniform(0,1)
    exp_rv = (-math.log(1-uniform_rv)/float(value))
    return exp_rv

def generate_nodes(N,A,D):
    
    #Generate nodes (computers/hosts) to test in our simulation
    nodes = []
    for i in range(0,N):
        nodes.append(Node(i*D, A))        
    return nodes

def csma_cd(N,A,R,L,D,S,persistent):
    
    #CSMA/CD SIMULATION
    
    packets_transmitted = 0
    packets_transmitted_success = 0
    
    current_time = 0
    nodes = generate_nodes(N,A,D)
    
    while True:
        #Pick the node having the smallest time (first packet to send on medium)
        
        first_node = Node(None,A)
        first_node.queue = [float("infinity")]
        
        #Note we use 'first_node.queue[0]' since we already sorted the packets...
        
        for my_node in nodes:
            if len(my_node.queue) > 0:
                first_node = first_node if first_node.queue[0] < my_node.queue[0] else my_node
        
        if first_node.position is None:
            break
        
        current_time = first_node.queue[0]
        packets_transmitted += 1
        
        #print(first_node.position)
        #break;
        
        collision_ctr = False
        
        for node in nodes:
            if node.position != first_node.position and len(node.queue) > 0:
                
                distance = abs(first_node.position - node.position)
                t_prop = distance/float(S)
                t_trans = L/float(R)
                
                if(node.queue[0] <= (current_time+t_prop)):
                    collision_flag = True
                else:
                    collision_flag = False
                
                if(current_time + t_prop)<node.queue[0]<(current_time+t_prop+t_trans):
                    if persistent is True:
                        for i in range(len(node.queue)):
                            if(current_time + t_prop)<node.queue[i]<(current_time+t_prop+t_trans):
                                node.queue[i] = (current_time+t_prop+t_trans)
                            else:
                                break
                    else:
                        node.medium_busy_np(R)
                       
                if collision_flag:
                    collision_ctr = True
                    packets_transmitted += 1
                    node.transmission_failed(R)
                    
        if collision_ctr is True:
            first_node.transmission_failed(R)
        else:
            packets_transmitted_success += 1
            first_node.drop_packet()
            
            
    print("Efficiency", packets_transmitted_success/float(packets_transmitted))
    print("Throughput", (L*packets_transmitted_success)/float(current_time+(L/R))*pow(10,-6), "Mbps")
    print("")
        

D = 10
C = 3*pow(10,8)
S = (2/float(3))*C

for N in range(20, 101, 20):
    for A in [5,12]:
        R = 1*pow(10,6)
        L = 1500
        print("Persistent", "Nodes: ",N, "Packets/s: ", A)
        csma_cd(N,A,R,L,D,S,True)
        
#for N in range(20, 101, 20):
#    for A in [7,10,20]:
#        R = 1*pow(10,6)
#        L = 1500
#        print("Non persistent", "Nodes: ",N, "Packets/s: ", A)
#        csma_cd(N,A,R,L,D,S,False)